import gzip
import logging
import multiprocessing
import os
import pathlib
import random
import shutil
from multiprocessing import Pool
from random import randrange

import re
from pathlib import Path
import ellc
import pandas as pd
import numpy as np
#import timesynth
from lcbuilder.star import starinfo
from lcbuilder.star.HabitabilityCalculator import HabitabilityCalculator
from lcbuilder.star.TicStarCatalog import TicStarCatalog

import astropy.units as u
import astropy.constants as ac
from lightkurve import TessLightCurve
from numpy.random import default_rng
from exoml.ml.physics.transit_functions import t14
from exoml.preparation.parser_utils import CreateTargetInput, create_star_csvs, detrend, prepare_lc_data


def create_injection_dataframe(injections_dir, lcs_dir):
    backeb_df = pd.read_csv(injections_dir + '/ete6_backeb_data.txt', comment="#")
    eb_df = pd.read_csv(injections_dir + '/ete6_eb_data.txt', comment="#")
    planet_df = pd.read_csv(injections_dir + '/ete6_planet_data.txt', comment="#")
    df = pd.DataFrame(columns=['object_id', 'type', 'period', 'epoch', 'Rsec/Rpri', 'b',
                                'a/Rstar', 'duration(h)', 'depth_primary', 'depth_secondary',
                               'insolation', 'Rstar_primary', 'Rstar_secondary', 'contact_amplitude', 'Mstar',
                               'radius(earth)'])
    i = 0
    log_count = 100
    random_number_generator = default_rng()
    for lc_file in list(pathlib.Path(lcs_dir).glob('*.fits')):
        object_id_no = int(str(lc_file).split('-')[1])
        object_id = 'TIC ' + str(object_id_no)
        # lc = TessLightCurve.read(lc_file)
        # object_id = lc.meta['OBJECT']
        star_file = lcs_dir + '/' + object_id + '_star.csv'
        star_df = pd.read_csv(star_file)
        star_radius = star_df['radius'].iloc[0]
        try:
            star_radius = 1 if np.isnan(star_radius) else star_radius
        except:
            print("ERROR WITH " + str(star_file))
        star_mass = star_df['mass'].iloc[0]
        try:
            if np.isnan(star_mass) and np.isnan(star_df['radius'].iloc[0]):
                star_mass = 1
            elif np.isnan(star_mass):
                star_mass = star_radius ** (1 / 0.8)
        except:
            print("ERROR WITH " + str(star_file))
        has_injected_object = False
        target_rows = backeb_df[backeb_df['TIC ID'] == object_id_no]
        for index, target_row in target_rows.iterrows():
            has_injected_object = True
            period = target_row['Orbital Period']
            epoch = target_row['Epoch [BTJD]']
            binary_rad = (target_row['Primary Star Depth'] * 1e-6 * (star_radius ** 2)) ** (1 / 2)
            if not np.isnan(target_row['Primary Rstar']):
                duration = t14(star_radius, star_mass, period, R_p=binary_rad)
            else:
                duration = t14(star_radius, star_mass, period)
            duration = duration * 24
            df = df.append({'object_id': object_id, 'type': 'bckEB', 'period': period,
                       'epoch': epoch,
                       'Rsec/Rpri': binary_rad / star_radius,
                       'b': target_row['Impact Parameter'],
                       'a/Rstar': 0, 'duration(h)': duration,
                       'depth_primary': target_row['Primary Star Depth'],
                       'depth_secondary': target_row['Primary Star Depth'],
                       'insolation': 0, 'Rstar_primary': target_row['Primary Rstar'],
                       'Rstar_secondary': target_row['Secondary Rstar'],
                       'contact_amplitude': target_row['Contact Binary Amplitude'],
                       'Mstar': star_mass,
                       'radius(earth)': (binary_rad * u.R_sun).to(u.R_earth).value}, ignore_index=True)
        target_rows = eb_df[eb_df['TIC ID'] == object_id_no]
        for index, target_row in target_rows.iterrows():
            has_injected_object = True
            period = target_row['Orbital Period']
            epoch = target_row['Epoch [BTJD]']
            if not np.isnan(target_row['Primary Rstar']):
                duration = t14(star_radius, star_mass, period, R_p=target_row['Primary Rstar'])
            else:
                duration = t14(star_radius, star_mass, period)
            duration = duration * 24
            binary_rad = (target_row['Primary Star Depth'] * 1e-6 * (star_radius ** 2)) ** (1 / 2)
            df = df.append({'object_id': object_id, 'type': 'EB', 'period': period,
                       'epoch': epoch,
                       'Rsec/Rpri': binary_rad / star_radius,
                       'b': target_row['Impact Parameter'],
                       'a/Rstar': 0, 'duration(h)': duration,
                       'depth_primary': target_row['Primary Star Depth'],
                       'depth_secondary': target_row['Primary Star Depth'],
                       'insolation': 0, 'Rstar_primary': target_row['Primary Rstar'],
                       'Rstar_secondary': target_row['Secondary Rstar'],
                       'contact_amplitude': target_row['Contact Binary Amplitude'],
                       'Mstar': star_mass,
                       'radius(earth)': (binary_rad * u.R_sun).to(u.R_earth).value}, ignore_index=True)
        target_rows = planet_df[planet_df['TIC ID'] == object_id_no]
        for index, target_row in target_rows.iterrows():
            has_injected_object = True
            period = target_row['Orbital Period']
            epoch = target_row['Epoch [BTJD]']
            duration = target_row['duration[h]']
            if np.isnan(duration) or duration == 0:
                duration = t14(star_radius, star_mass, period)
                duration = duration * 24
            df = df.append({'object_id': object_id, 'type': 'planet', 'period': period,
                       'epoch': epoch,
                       'Rsec/Rpri': target_row['Rp/Rstar'],
                       'b': target_row['Impact Parameter'],
                       'a/Rstar': target_row['a/Rstar'], 'duration(h)': duration,
                       'depth_primary': target_row['depth'],
                       'depth_secondary': 0,
                       'insolation': target_row['Insolation Flux'],
                       'Rstar_primary': target_row['Rstar'],
                       'Rstar_secondary': target_row['Rstar'] * target_row['Rp/Rstar'],
                       'contact_amplitude': 0,
                       'Mstar': star_mass,
                       'radius(earth)': target_row['Rp']}, ignore_index=True)
        while not has_injected_object:
            min_period = 0.5
            max_period = 10
            period = random_number_generator.uniform(min_period, max_period)
            depth_ppm = random_number_generator.uniform(100, 500)
            planet_radius = (depth_ppm / 1e6 * ((star_radius * u.R_sun).to(u.R_earth).value ** 2)) ** 0.5
            # min_lc_time = 1612.85 + 0.5
            # epoch = random_number_generator.uniform(min_lc_time, min_lc_time + period)
            # duration = t14(star_radius, star_mass, period)
            # if 2 * duration < period:
            #     duration = duration * 24
            #     df = df.append({'object_id': object_id, 'type': 'none', 'period': period,
            #                     'epoch': epoch,
            #                     'Rsec/Rpri': 1e-7,
            #                     'b': 0,
            #                     'a/Rstar': 1e-7, 'duration(h)': duration,
            #                     'depth_primary': depth_ppm,
            #                     'depth_secondary': 0,
            #                     'insolation': 0, 'Rstar_primary': 1e-7,
            #                     'Rstar_secondary': 1e-7,
            #                     'contact_amplitude': 0,
            #                     'Mstar': star_mass,
            #                     'radius(earth)': planet_radius}, ignore_index=True)
            #     has_injected_object = True
            for match in Path("/data/scratch/ml/ete6/lcs/").glob(object_id + '_*_input_scalar_values.csv'):
                period = float(re.findall("\d+\.\d+", str(match))[0])
                min_lc_time = 1612.85 + 0.5
                epoch = random_number_generator.uniform(min_lc_time, min_lc_time + period)
                duration = t14(star_radius, star_mass, period)
                while 2 * duration >= period:
                    duration = duration / 2
                if 2 * duration < period:
                    duration = duration * 24
                    df = df.append({'object_id': object_id, 'type': 'none', 'period': period,
                                    'epoch': epoch,
                                    'Rsec/Rpri': 1e-7,
                                    'b': 0,
                                    'a/Rstar': 1e-7, 'duration(h)': duration,
                                    'depth_primary': depth_ppm,
                                    'depth_secondary': 0,
                                    'insolation': 0, 'Rstar_primary': 1e-7,
                                    'Rstar_secondary': 1e-7,
                                    'contact_amplitude': 0,
                                    'Mstar': star_mass,
                                    'radius(earth)': planet_radius}, ignore_index=True)
                    has_injected_object = True
                    break
                else:
                    print("inloop")
        print("Processed " + str(i) + " TICs.")
        if (i + 1) % log_count == 0:
            print("Processed " + str(i) + " TICs.")
            #df = df.sort_values(["TIC ID", "type"], ascending=True)
            #df.to_csv(injections_dir + '/injected_objects.csv')
        i = i + 1
    df = df.sort_values(["object_id", "type"], ascending=True)
    df.to_csv(injections_dir + '/injected_objects.csv')


def create_target_csvs(lcs_dir, models_dir, cores=multiprocessing.cpu_count() - 1, max_lc_length=20610, force=False,
                       ids=None, mode=['all']):
    injected_files_df = pd.read_csv(models_dir + '/injected_objects.csv')
    inputs = []
    if ids is not None:
        injected_files_df = injected_files_df[injected_files_df['object_id'].isin(ids)]
    injected_files_df = injected_files_df.sort_values(by=["period"], ascending=True)
    injected_files_df = injected_files_df.reset_index(drop=True)
    ranges = range(0, len(injected_files_df))
    print("Total number of targets is " + str(len(injected_files_df)))
    for index in ranges:
        target_row = injected_files_df.iloc[index]
        object_id = target_row['object_id']
        mission_id = object_id.split(' ')[1]
        lc_file = lcs_dir + "/tess2019128220341-" + str(mission_id).zfill(16) + "-0016-s_lc.fits"
        tpf_file = lcs_dir + "/tpfs/tess2019128220341-" + str(mission_id).zfill(16) + "-0016-s_tp.fits"
        inputs = inputs + [CreateTargetInput('TIC', object_id, lc_file, lcs_dir, models_dir, target_row,
                                             None, index, max_lc_length, tpf_files=[tpf_file], force=force, mode=mode)]
    with Pool(processes=cores, maxtasksperchild=1) as pool:
        pool.map(_create_target_csv, inputs, chunksize=1)

def create_injected_target_csvs(lcs_dir, models_dir, cores=multiprocessing.cpu_count() - 1,
                                max_lc_length=20610, force=False, ids=None):
    injected_files_df = pd.read_csv(models_dir + '/injected_objects.csv')
    injected_files_df = injected_files_df[injected_files_df['type'] == 'none']
    inputs = []
    if ids is not None:
        injected_files_df = injected_files_df[injected_files_df['object_id'].isin(ids)]
    for index, target_row in injected_files_df.iterrows():
        object_id = target_row['object_id']
        mission_id = object_id.split(' ')[1]
        lc_file = lcs_dir + "/tess2019128220341-" + str(mission_id).zfill(16) + "-0016-s_lc.fits"
        tpf_file = lcs_dir + "/tpfs/tess2019128220341-" + str(mission_id).zfill(16) + "-0016-s_tp.fits"
        inputs = inputs + [CreateTargetInput('TIC', object_id, lc_file, lcs_dir, models_dir, target_row,
                                             None, max_lc_length, tpf_files=[tpf_file], force=force,
                                             inject_random=True)]
    with Pool(processes=cores, maxtasksperchild=1) as pool:
        pool.map(_create_target_csv, inputs, chunksize=1)

def _create_target_csv(create_target_input):
    pid = multiprocessing.current_process().pid
    type_dict = {0: 'none', 1: 'bckEB', 2: 'EB', 3: 'planet'}
    file_prefix = create_target_input.target_dir + '/' + create_target_input.id
    file_planet_prefix = file_prefix + '_' + \
                         str(round(create_target_input.injected_files_df['period'], 2))
    logging.info("PID %s: Trying to get info for %s", pid, file_planet_prefix)
    file_name = file_planet_prefix + '_metrics_short.csv'
    if create_target_input.force or not os.path.exists(file_name):
        try:
            lc_df = pd.DataFrame(columns=['#time', 'flux', 'flux_err', 'planet_model', 'eb_model', 'bckeb_model',
                                          'centroid_ra', 'centroid_dec', 'og_flux', 'bck_flux'])
            lc = TessLightCurve.read(create_target_input.lc_file)
            mission_id = lc.meta['OBJECT']
            mission_id = int(mission_id.split(' ')[1])
            leading_zeros_object_id = '{:09}'.format(mission_id)
            lc_length = len(lc)
            eb_model_flux = np.ones(lc_length)
            backeb_model_flux = np.ones(lc_length)
            planet_model_flux = np.ones(lc_length)
            model_file = create_target_input.models_dir + '/ebs/EBs/EBs' + '_' + leading_zeros_object_id + '.txt'
            star_info = starinfo.StarInfo(create_target_input.id, *TicStarCatalog().catalog_info(mission_id))
            if create_target_input.inject_random:
                type_random = randrange(4)
                type = type_dict[type_random]
                rstar = star_info.radius * u.Rsun
                time = lc.time.value
                flux = lc.flux.value
                period = create_target_input.injected_files_df['period']
                t0 = create_target_input.injected_files_df['epoch']
                texpo = 120 / 60. / 60. / 24.
                if type == 'none':
                    a = 1
                elif type == 'bckEB':
                    rtransit = random.uniform(0.1, rstar / 2) * u.Rsun
                    a = HabitabilityCalculator().calculate_semi_major_axis(period, star_info.mass)
                    model = lc(
                        t_obs=time,
                        radius_1=rstar.to(u.au) / a,  # star radius convert from AU to in units of a
                        radius_2=rtransit.to(u.au) / a,
                        # convert from Rearth (equatorial) into AU and then into units of a
                        sbratio=0,
                        incl=90,
                        light_3=0,
                        t_zero=t0,
                        period=period,
                        a=None,
                        q=1e-6,
                        f_c=None, f_s=None,
                        ldc_1=[0.5, 0.5], ldc_2=None,
                        gdc_1=None, gdc_2=None,
                        didt=None,
                        domdt=None,
                        rotfac_1=1, rotfac_2=1,
                        hf_1=1.5, hf_2=1.5,
                        bfac_1=None, bfac_2=None,
                        heat_1=None, heat_2=None,
                        lambda_1=None, lambda_2=None,
                        vsini_1=None, vsini_2=None,
                        t_exp=texpo, n_int=None,
                        grid_1='default', grid_2='default',
                        ld_1='quad', ld_2=None,
                        shape_1='sphere', shape_2='sphere',
                        spots_1=None, spots_2=None,
                        exact_grav=False, verbose=1)
                    model = model / 10
                    planet_model_flux = flux + model - 1.
                elif type == 'EB':
                    rtransit = random.uniform(0.1, rstar / 2) * u.Rsun
                    a = HabitabilityCalculator().calculate_semi_major_axis(period, star_info.mass)
                    model = lc(
                        t_obs=time,
                        radius_1=rstar.to(u.au) / a,  # star radius convert from AU to in units of a
                        radius_2=rtransit.to(u.au) / a,
                        # convert from Rearth (equatorial) into AU and then into units of a
                        sbratio=0,
                        incl=90,
                        light_3=0,
                        t_zero=t0,
                        period=period,
                        a=None,
                        q=1e-6,
                        f_c=None, f_s=None,
                        ldc_1=[0.5, 0.5], ldc_2=None,
                        gdc_1=None, gdc_2=None,
                        didt=None,
                        domdt=None,
                        rotfac_1=1, rotfac_2=1,
                        hf_1=1.5, hf_2=1.5,
                        bfac_1=None, bfac_2=None,
                        heat_1=None, heat_2=None,
                        lambda_1=None, lambda_2=None,
                        vsini_1=None, vsini_2=None,
                        t_exp=texpo, n_int=None,
                        grid_1='default', grid_2='default',
                        ld_1='quad', ld_2=None,
                        shape_1='sphere', shape_2='sphere',
                        spots_1=None, spots_2=None,
                        exact_grav=False, verbose=1)
                    planet_model_flux = flux + model - 1.
                elif type == 'planet':
                    curve_rms = np.std(flux)
                    rms_in_radius = (curve_rms * (star_info.radius ** 2)) ** (1 / 2)
                    rtransit = random.uniform(rms_in_radius / 3, 0.1) * u.Rsun
                    a = HabitabilityCalculator().calculate_semi_major_axis(period, star_info.mass)
                    model = lc(
                        t_obs=time,
                        radius_1=rstar.to(u.au) / a,  # star radius convert from AU to in units of a
                        radius_2=rtransit.to(u.au) / a,
                        # convert from Rearth (equatorial) into AU and then into units of a
                        sbratio=0,
                        incl=90,
                        light_3=0,
                        t_zero=t0,
                        period=period,
                        a=None,
                        q=1e-6,
                        f_c=None, f_s=None,
                        ldc_1=[0.5, 0.5], ldc_2=None,
                        gdc_1=None, gdc_2=None,
                        didt=None,
                        domdt=None,
                        rotfac_1=1, rotfac_2=1,
                        hf_1=1.5, hf_2=1.5,
                        bfac_1=None, bfac_2=None,
                        heat_1=None, heat_2=None,
                        lambda_1=None, lambda_2=None,
                        vsini_1=None, vsini_2=None,
                        t_exp=texpo, n_int=None,
                        grid_1='default', grid_2='default',
                        ld_1='quad', ld_2=None,
                        shape_1='sphere', shape_2='sphere',
                        spots_1=None, spots_2=None,
                        exact_grav=False, verbose=1)
                    planet_model_flux = flux + model - 1.
            if os.path.exists(model_file):
                model_df = pd.read_csv(model_file, header=None)
                eb_model_flux = model_df[0].to_numpy()
                eb_model_flux = eb_model_flux / np.median(eb_model_flux)
                eb_model_flux = resample_model(eb_model_flux, lc.time.value)
            model_file = create_target_input.models_dir + '/backebs/BackEBs/BackEBs_' + leading_zeros_object_id + '.txt'
            if os.path.exists(model_file):
                model_df = pd.read_csv(model_file, header=None)
                backeb_model_flux = model_df[0].to_numpy()
                backeb_model_flux = backeb_model_flux / np.median(backeb_model_flux)
                backeb_model_flux = resample_model(backeb_model_flux, lc.time.value)
                # import matplotlib.pyplot as plt
                # plt.scatter(np.linspace(lc.time.value[0], lc.time.value[-1], len(eb_model_flux)), eb_model_flux)
                # plt.scatter(lc.time.value, lc.pdcsap_flux.value / np.median(lc.pdcsap_flux.value))
                # plt.show()
            model_file = create_target_input.models_dir + '/planets/Planets/Planets' + '_' + leading_zeros_object_id + '.txt'
            if os.path.exists(model_file):
                model_df = pd.read_csv(model_file, header=None)
                planet_model_flux = model_df[0].to_numpy()
                planet_model_flux = planet_model_flux / np.median(planet_model_flux)
                planet_model_flux = resample_model(planet_model_flux, lc.time.value)
            padding_zeros_count = create_target_input.max_lc_length - lc_length
            padding_zeros_count = 0 if padding_zeros_count < 0 else padding_zeros_count
            padding_zeros_count_left = padding_zeros_count // 2 + 1 if padding_zeros_count % 2 != 0 else padding_zeros_count // 2
            padding_zeros_count_right = padding_zeros_count // 2
            padding_zeros_left = np.zeros(padding_zeros_count_left)
            padding_zeros_right = np.zeros(padding_zeros_count_right)
            padding_ones_left = np.zeros(padding_zeros_count_left)
            padding_ones_right = np.zeros(padding_zeros_count_right)
            flux_values = np.array(lc.pdcsap_flux.value)
            median_flux = np.median(flux_values)
            flux_values = flux_values / median_flux
            if len(np.argwhere(flux_values < 0).flatten()) > 0:
                flux_values = np.array(lc.pdcsap_flux.value) + 2 * np.abs(np.nanmin(lc.pdcsap_flux.value))
                median_flux = np.median(flux_values)
                flux_values = flux_values / median_flux
            time_values = lc.time.value
            duration = create_target_input.injected_files_df['duration(h)'] / 24
            duration_to_period = duration / create_target_input.injected_files_df['period']
            flux_detrend_0 = detrend(time_values, flux_values, duration_to_period * 4)
            lc_df['flux_0'] = flux_detrend_0
            bck_flux_detrend = detrend(time_values, flux_values, duration_to_period * 4, smooth=False)
            lc_df['background_flux'] = bck_flux_detrend
            flux_err_values = np.array(lc.flux_err.value) / median_flux
            lc_df['time'] = np.array(lc.time.value)
            lc_df['flux'] = flux_values
            lc_df['flux_err'] = flux_err_values
            lc_df['centroids_x'] = np.array(lc.mom_centr1.value)
            lc_df['centroids_y'] = np.array(lc.mom_centr2.value)
            lc_df['motion_x'] = np.array(lc.pos_corr1.value)
            lc_df['motion_y'] = np.array(lc.pos_corr2.value)
            lc_df['eb_model'] = np.array(eb_model_flux)
            lc_df['bckeb_model'] = np.array(backeb_model_flux)
            lc_df['planet_model'] = np.array(planet_model_flux)
            lc_df = lc_df.fillna(0)
            prepare_lc_data(file_planet_prefix, create_target_input.id, lc_df,
                            create_target_input.injected_files_df['duration(h)'] / 24,
                            create_target_input.injected_files_df['period'],
                            create_target_input.injected_files_df['epoch'],
                            star_info.ra, star_info.dec, star_info, False, tpf_files=create_target_input.tpf_files,
                            modes=create_target_input.mode)
        except Exception as e:
            logging.exception("PID %s: Failed get info for %s", pid, file_planet_prefix)


def create_injected_target_csvs(create_target_input):
    random_number_generator = default_rng()
    lc_df = pd.DataFrame(columns=['#time', 'flux', 'flux_err',
                                  'centroid_x', 'centroid_y', 'motion_x', 'motion_y', 'bck_flux'])
    lc_df_injected = pd.DataFrame(columns=['#time', 'flux', 'flux_err', 'planet_model',
                                  'centroid_x', 'centroid_y', 'motion_x', 'motion_y', 'bck_flux'])
    try:
        lc = TessLightCurve.read(create_target_input.lc_file)
    except Exception as e:
        print("Invalid fits file " + create_target_input.lc_file)
        return
    try:
        star_df = pd.read_csv(create_target_input.star_file)
    except Exception as e:
        print("Invalid fits file " + create_target_input.lc_file)
        return
    star_radius = star_df.loc[0, 'radius']
    star_mass = star_df.loc[0, 'mass']
    ld = (star_df.loc[0, 'ld_a'], star_df.loc[0, 'ld_b'])
    object_id = lc.meta['OBJECT']
    object_id = int(object_id.split(' ')[1])
    mission_id = 'TIC'
    injected_file_row = create_target_input.injected_files_df[
        create_target_input.injected_files_df['TIC ID'] == object_id]
    leading_zeros_object_id = '{:09}'.format(object_id)
    lc_length = len(lc)
    padding_zeros_count = create_target_input.max_lc_length - lc_length
    padding_zeros_count = 0 if padding_zeros_count < 0 else padding_zeros_count
    padding_zeros_count_left = padding_zeros_count // 2 + 1 if padding_zeros_count % 2 != 0 else padding_zeros_count // 2
    padding_zeros_count_right = padding_zeros_count // 2
    padding_zeros_left = np.zeros(padding_zeros_count_left)
    padding_zeros_right = np.zeros(padding_zeros_count_right)
    padding_ones_left = np.zeros(padding_zeros_count_left)
    padding_ones_right = np.zeros(padding_zeros_count_right)
    planet_model_flux = np.ones(lc_length)
    planets_num = random_number_generator.uniform(1, 3)
    planet_periods = np.zeros(planets_num)
    planet_t0s = np.zeros(planets_num)
    planet_radius = np.zeros(planets_num)
    rms = np.std(lc.flux)
    rms_ppm = np.std(lc.flux) * 1000000 # ppms
    times = lc.time.value
    flux = lc.flux.value
    modified_flux = flux
    max_period = 0.5
    red_noise_number = random_number_generator.uniform(0, 3)
    for red_noise in red_noise_number:
        red_noise_freq = random_number_generator.uniform(0.01, 2)
        flux_std = rms
        red_noise_sd = random_number_generator.uniform(0, flux_std * 3)
        red_noise = timesynth.noise.RedNoise(std=red_noise_sd, tau=red_noise_freq)
        timeseries_rn = np.zeros(len(times))
        timeseries_rn = []
        for value in times:
            rn_value = red_noise.sample_next(value, None, None)
            rn_value = rn_value[0] if isinstance(rn_value, (list, np.ndarray)) else rn_value
            timeseries_rn = timeseries_rn + [rn_value]
        timeseries_rn = np.array(timeseries_rn)
        modified_flux = modified_flux + timeseries_rn
    #TODO write flux with red noise
    for planet in np.arange(0, planets_num):
        planet_periods[planet] = random_number_generator.uniform(0.5, 20)
        max_period = max_period if planet_periods[planet] < max_period else planet_periods[planet]
        planet_t0s[planet] = lc.time[0] + planet_periods[planet] / random_number_generator.uniform(0, planet_periods[planet])
        num_transits_pesimistic = planet_t0s[planet] // len(lc.time.value)
        depth_ppm = random_number_generator(rms_ppm // num_transits_pesimistic, rms_ppm * 4)
        planet_radius[planet] = np.sqrt((depth_ppm / 1e6) * (star_radius ** 2))
        P1 = planet_periods[planet] * u.day
        a = np.cbrt((ac.G * star_mass * u.M_sun * P1 ** 2) / (4 * np.pi ** 2)).to(u.au)
        planet_model = ellc.lc(
            t_obs=times,
            radius_1=(star_radius * u.R_sun).to(u.au) / a,  # star radius convert from AU to in units of a
            radius_2=(planet_radius * u.R_earth).to(u.au) / a,
            # convert from Rearth (equatorial) into AU and then into units of a
            sbratio=0,
            incl=90,
            light_3=0,
            t_zero=planet_t0s[planet],
            period=planet_periods[planet],
            a=None,
            q=1e-6,
            f_c=None, f_s=None,
            ldc_1=ld, ldc_2=None,
            gdc_1=None, gdc_2=None,
            didt=None,
            domdt=None,
            rotfac_1=1, rotfac_2=1,
            hf_1=1.5, hf_2=1.5,
            bfac_1=None, bfac_2=None,
            heat_1=None, heat_2=None,
            lambda_1=None, lambda_2=None,
            vsini_1=None, vsini_2=None,
            t_exp=None, n_int=None,
            grid_1='default', grid_2='default',
            ld_1='quad', ld_2=None,
            shape_1='sphere', shape_2='sphere',
            spots_1=None, spots_2=None,
            exact_grav=False, verbose=1)
        modified_flux = modified_flux - planet_model + 1
    bck_flux_detrend = detrend(times, lc.background_flux.value, 1, smooth=False)
    lc_df['bck_flux'] = np.concatenate((padding_ones_left, bck_flux_detrend, padding_ones_right))
    flux_err_values = np.array(lc.flux_err.value) / np.nanmedian(modified_flux)
    lc_df['#time'] = np.concatenate((padding_zeros_left, np.array(lc.time.value), padding_zeros_right))
    lc_df['flux'] = np.concatenate((padding_ones_left, modified_flux, padding_ones_right))
    lc_df['flux_err'] = np.concatenate((padding_zeros_left, flux_err_values, padding_zeros_right))
    lc_df['centroid_x'] = np.concatenate((padding_zeros_left, np.array(lc.centroid_col.value), padding_zeros_right))
    lc_df['centroid_y'] = np.concatenate((padding_zeros_left, np.array(lc.centroid_row.value), padding_zeros_right))
    lc_df['motion_x'] = np.concatenate((padding_zeros_left, np.array(lc.mom_centr1.value), padding_zeros_right))
    lc_df['motion_y'] = np.concatenate((padding_zeros_left, np.array(lc.mom_centr2.value), padding_zeros_right))
    lc_df['planet_model'] = np.concatenate((padding_ones_left, np.array(planet_model_flux), padding_ones_right))
    lc_df = lc_df.fillna(0)
    lc_df.to_csv(lcs_dir + '/' + mission_id + '_' + str(object_id) + '_noise_lc.csv')
    #TODO write final injected flux
    padding_zeros_count = create_target_input.max_lc_length - lc_length
    padding_zeros_count = 0 if padding_zeros_count < 0 else padding_zeros_count
    padding_zeros_count_left = padding_zeros_count // 2 + 1 if padding_zeros_count % 2 != 0 else padding_zeros_count // 2
    padding_zeros_count_right = padding_zeros_count // 2
    padding_zeros_left = np.zeros(padding_zeros_count_left)
    padding_zeros_right = np.zeros(padding_zeros_count_right)
    padding_ones_left = np.zeros(padding_zeros_count_left)
    padding_ones_right = np.zeros(padding_zeros_count_right)
    flux_values = np.array(lc.pdcsap_flux.value)
    median_flux = np.nanmedian(flux_values)
    flux_values = flux_values / median_flux
    if len(np.argwhere(flux_values < 0).flatten()) > 0:
        flux_values = np.array(lc.pdcsap_flux.value) + 2 * np.abs(np.nanmin(lc.pdcsap_flux.value))
        median_flux = np.median(flux_values)
        flux_values = flux_values / median_flux
    time_values = lc.time.value
    duration = t14(star_radius, star_mass, max_period)
    duration_to_period = duration / max_period
    bck_flux_detrend = detrend(time_values, lc.background_flux.value, duration_to_period * 4, smooth=False)
    lc_df['bck_flux'] = np.concatenate((padding_ones_left, bck_flux_detrend, padding_ones_right))
    flux_err_values = np.array(lc.flux_err.value) / median_flux
    lc_df['#time'] = np.concatenate((padding_zeros_left, np.array(lc.time.value), padding_zeros_right))
    lc_df['flux'] = np.concatenate((padding_ones_left, flux_values, padding_ones_right))
    lc_df['flux_err'] = np.concatenate((padding_zeros_left, flux_err_values, padding_zeros_right))
    lc_df['centroid_x'] = np.concatenate((padding_zeros_left, np.array(lc.centroid_col.value), padding_zeros_right))
    lc_df['centroid_y'] = np.concatenate((padding_zeros_left, np.array(lc.centroid_row.value), padding_zeros_right))
    lc_df['motion_x'] = np.concatenate((padding_zeros_left, np.array(lc.mom_centr1.value), padding_zeros_right))
    lc_df['motion_y'] = np.concatenate((padding_zeros_left, np.array(lc.mom_centr2.value), padding_zeros_right))
    lc_df['planet_model'] = np.concatenate((padding_ones_left, np.array(planet_model_flux), padding_ones_right))
    lc_df = lc_df.fillna(0)
    lc_df.to_csv(lcs_dir + '/' + mission_id + '_' + str(object_id) + '_injected_lc.csv')


def resample_model(model, lc_time):
    total_missing_points = len(model) - len(lc_time)
    gaps_sizes = []
    time_gaps_indexes = np.argwhere(np.abs(lc_time[1:] - lc_time[:-1]) > 0.014).flatten()
    for time_gap_index in time_gaps_indexes:
        gaps_sizes = gaps_sizes + [lc_time[time_gap_index + 1] - lc_time[time_gap_index]]
    total_gap_size = np.sum(gaps_sizes)
    gap_sizes_percent = [gap_size / total_gap_size for gap_size in gaps_sizes]
    gap_points = []
    for gap_size_percent in gap_sizes_percent[:-1]:
        gap_points = gap_points + [int(np.round(total_missing_points * gap_size_percent))]
    gap_points = gap_points + [total_missing_points - np.sum(gap_points)]
    added_gap_points = 0
    i = 0
    corrected_time_gap_indexes = np.array([])
    for time_gap_index in time_gaps_indexes:
        corrected_time_gap_index = time_gap_index + added_gap_points
        corrected_time_gap_indexes = np.concatenate((corrected_time_gap_indexes, np.arange(corrected_time_gap_index + 1, corrected_time_gap_index + 1 + gap_points[i])))
        added_gap_points = gap_points[i]
        i = i + 1
    model = np.delete(model, corrected_time_gap_indexes.astype(int))
    return model


def uncompress_data(data_dir):
    for file in os.listdir(data_dir):
        original_file_name = data_dir + '/' + file
        destination_file_name = file.split('.')[0]
        destination_file_name = data_dir + '/' + destination_file_name + '.txt'
        with gzip.open(original_file_name, 'rb') as f_in:
            with open(destination_file_name, 'wb') as f_out:
                shutil.copyfileobj(f_in, f_out)



import glob
import logging
import multiprocessing
import os
import sys
import pathlib
import re
from itertools import chain
from multiprocessing import Pool

import gzip
import shutil
import lcbuilder.constants
#import triceratops.triceratops as tr
import astropy.units as u
import foldedleastsquares
import lightkurve
import numpy as np
import wotan
import pandas as pd
from lcbuilder import constants
from astropy.coordinates import SkyCoord
from astropy.wcs import WCS
from astroquery.mast import Catalogs, TesscutClass
from lcbuilder.helper import LcbuilderHelper
from lcbuilder.lcbuilder_class import LcBuilder
from lcbuilder.objectinfo.MissionInputObjectInfo import MissionInputObjectInfo
from lcbuilder.objectinfo.MissionObjectInfo import MissionObjectInfo
from lcbuilder.objectinfo.preparer.MissionInputLightcurveBuilder import MissionInputLightcurveBuilder
from lcbuilder.objectinfo.preparer.MissionLightcurveBuilder import MissionLightcurveBuilder
from lcbuilder.photometry.aperture_extractor import ApertureExtractor
from lcbuilder.star.HabitabilityCalculator import HabitabilityCalculator
from lcbuilder.star.starinfo import StarInfo
from lightkurve import TessLightCurve, KeplerLightCurve, LightCurveCollection
from uncertainties import ufloat
from watson.watson import Watson


def detrend(time_array, flux_array, detrend_window, time_threshold=0.0001, smooth=True, smooth_window_len=11,
            window='blackman'):
    flux_without_zeros = flux_array[time_array > time_threshold]
    time_without_zeros = time_array[time_array > time_threshold]
    if smooth:
        if window not in ['flat', 'hanning', 'hamming', 'bartlett', 'blackman']:
            raise ValueError("Window is on of 'flat', 'hanning', 'hamming', 'bartlett', 'blackman'")
        s = np.r_[np.ones(smooth_window_len // 2), flux_without_zeros, np.ones(smooth_window_len // 2)]
        # print(len(s))
        if smooth_window_len >= 3:
            if window == 'flat':  # moving average
                w = np.ones(smooth_window_len, 'd')
            else:
                w = eval('np.' + window + '(smooth_window_len)')
            flux_without_zeros = np.convolve(w / w.sum(), s, mode='valid')
    flux_without_zeros = np.nan_to_num(wotan.flatten(time_without_zeros, flux_without_zeros, detrend_window,
                                                     method="biweight"))
    flux_array[time_array > time_threshold] = flux_without_zeros
    return flux_array


class CreateStarCsvInput:
    def __init__(self, object_id, pixel_size, search_radius, star_catalog, output_dir, force) -> None:
        super().__init__()
        self.object_id = object_id
        self.search_radius = search_radius
        self.pixel_size = pixel_size * u.arcsec
        self.star_catalog = star_catalog
        self.output_dir = output_dir
        self.force = force

def create_star_csvs(lcs_dir, search_radius, pixel_size, star_catalog, csv,
                     cores=os.cpu_count() - 1, ids=None, force=False):
    tces_df = pd.read_csv(csv, comment='#')
    tces_df = tces_df.sort_values(by=["object_id"], ascending=True)
    tces_df = tces_df.drop_duplicates(subset=["object_id"], keep='last')
    tces_df = tces_df.sample(frac=1).reset_index(drop=True)
    print("Total number of targets is " + str(len(tces_df)))
    create_star_inputs = []
    if ids is not None:
        tces_df = tces_df[tces_df['object_id'].isin(ids)]
    for index, tce_row in tces_df.iterrows():
        create_star_inputs = create_star_inputs + [CreateStarCsvInput(tce_row['object_id'], pixel_size,
                                                                      search_radius, star_catalog, lcs_dir,
                                                                      force)]
    with Pool(processes=cores, maxtasksperchild=1) as pool:
        pool.map(create_star_csv, create_star_inputs, chunksize=1)

def create_star_csv(create_star_input):
    tries = 0
    star_file = create_star_input.output_dir + '/' + create_star_input.object_id + '_star.csv'
    mission, mission_prefix, mission_id = LcBuilder().parse_object_info(create_star_input.object_id)
    if not os.path.exists(star_file) or create_star_input.force:
        while tries < 3:
            try:
                star_info = create_star_input.star_catalog.catalog_info(mission_id)
                star_df = pd.DataFrame(
                    columns=['id', 'ra', 'dec', 'ld_a', 'ld_b', 'Teff', 'lum', 'logg', 'radius', 'mass', 'v', 'j', 'h',
                             'k',
                             'dist_arcsec'])
                star_df = pd.concat([star_df, pd.DataFrame.from_dict({'id': [create_star_input.object_id], 'ra': [star_info[11]], 'dec': [star_info[12]],
                                          'ld_a': [star_info[0][0]],
                                          'ld_b': [star_info[0][1]], 'Teff': [star_info[1]],
                                          'lum': [star_info[2]], 'logg': [star_info[3]], 'radius': [star_info[5]],
                                          'mass': [star_info[8]], 'v': [star_info[13]], 'j': [star_info[15]],
                                          'h': [star_info[17]],
                                          'k': [star_info[19]], 'dist_arcsec': [0]}, orient='columns')], ignore_index=True)
                break
            except Exception as e:
                logging.exception("Failed object %s try", create_star_input.object_id)
                tries = tries + 1
        if tries >= 3:
            print("Failed downloading object id " + str(create_star_input.object_id))
        assert tries < 3
        try:  # Acquiring neighbours parameters
            ra = star_df['ra'].iloc[0]
            dec = star_df['dec'].iloc[0]
            if ra is not None and dec is not None:
                ticid = Catalogs.query_region(
                    SkyCoord(ra, dec, unit="deg"),
                    radius=(create_star_input.search_radius * (create_star_input.pixel_size / 3600)) * u.deg,
                    catalog="TIC"
                )[0]["ID"]
            df = Catalogs.query_object(
                "TIC" + str(ticid),
                radius=(create_star_input.search_radius * (create_star_input.pixel_size / 3600)) * u.deg,
                catalog="TIC"
            )
            stars = df.to_pandas()
            sep = [0]
            pa = [0]
            c_target = SkyCoord(
                stars["ra"].values[0],
                stars["dec"].values[0],
                unit="deg"
            )
            for i in range(1, len(stars)):
                c_star = SkyCoord(
                    stars["ra"].values[i],
                    stars["dec"].values[i],
                    unit="deg"
                )
                sep.append(
                    np.round(
                        c_target.separation(c_star).to(u.arcsec).value,
                        3
                    )
                )
                pa.append(
                    np.round(
                        c_target.position_angle(c_star).to(u.deg).value,
                        3
                    )
                )
            stars["dist_arcsec"] = sep
            stars["PA (E of N)"] = pa  # TODO should we use this?
            for index, star in stars.iterrows():
                star_df = pd.concat([star_df, pd.DataFrame.from_dict({'id': [star['ID']], 'ra': [star['ra']], 'dec': [star['dec']], 'ld_a': [0],
                                          'ld_b': [0], 'Teff': [star['Teff']],
                                          'lum': [star['lum']], 'logg': [star['logg']], 'radius': [star['rad']],
                                          'mass': [star['mass']], 'v': [star['Vmag']], 'j': [star['Jmag']],
                                          'h': [star['Hmag']],
                                          'k': [star['Kmag']], 'dist_arcsec': [star['dist_arcsec']]}, orient='columns')], ignore_index=True)
            star_df = star_df.sort_values(["dist_arcsec"], ascending=True)
        except Exception as e:
            logging.exception('Something failed when retrieving neighbours info for object id %s',
                              str(create_star_input.object_id))
        star_df.to_csv(star_file)
    logging.info('Processed object id ' + str(create_star_input.object_id))

def prepare_lc_data(file_prefix, object_id, lc_data, transit_duration, transit_period, epoch, ra, dec, star_info: StarInfo,
                    authors: list[str], cadences: list[int], cadence_name, tpf_files=None, modes=['all'], sectors='all'):
    lc_df = None
    if 'quality' in lc_data.columns:
        lc_data = lc_data.drop(['quality'], axis=1)
    lc_data = lc_data.loc[~lc_data['flux'].isna()]
    lcbuilder = LcBuilder()
    mission, mission_prefix, id = lcbuilder.parse_object_info(object_id)
    if tpf_files is None:
        tpfs = lightkurve.search_targetpixelfile(object_id, sector=sectors if isinstance(sectors, list) else None,
                                                 campaign=sectors if isinstance(sectors, list)  else None,
                                                 quarter=sectors if isinstance(sectors, list)  else None,
                                                 exptime=cadences, author=authors).download_all()
    else:
        tpfs = []
        for file in tpf_files:
            tpfs.append(lightkurve.TessTargetPixelFile(file, ignore_missing_simple=True))
    lc_filename = str(file_prefix) + '_lc_' + str(cadence_name) + '.csv'
    transit_duration = transit_duration if transit_duration < transit_period else transit_period / 2
    duration_to_period = transit_duration / transit_period
    if 'all' in modes or 'lc' in modes:
        lc_df = pd.DataFrame(columns=['#time', 'flux', 'flux_err', 'planet_model', 'eb_model', 'bckeb_model',
                                      'centroid_x', 'centroid_y', 'motion_x', 'motion_y', 'bck_flux'])
        flux_values = np.array(lc_data.flux)
        # Normalizing PDCSAP flux, but not using the normalization for detrending
        median_flux = np.nanmedian(flux_values)
        flux_values = flux_values / median_flux
        if len(np.argwhere(flux_values < 0).flatten()) > 0:
            flux_values = np.array(lc_data.flux)
            flux_values = flux_values + 2 * np.abs(np.nanmin(flux_values))
            median_flux = np.nanmedian(flux_values)
            flux_values = flux_values / median_flux
        time_values = np.array(lc_data.time)
        series_len = len(time_values)
        median_cadence = LcbuilderHelper.compute_cadence(time_values) / 3600 / 24
        for j in np.arange(0, 1, 1):
            detrend_window = transit_duration * (4 + j)
            logging.info("Detrending no %s for target %s with %s points with duration %s", j, object_id, series_len, detrend_window)
            if detrend_window > median_cadence * 2:
                flux_detrend_j = detrend(time_values, np.array(lc_data.flux), detrend_window, smooth=False)
            else:
                flux_detrend_j = np.array(lc_data.flux)
            lc_df['flux_' + str(j)] = flux_detrend_j
        logging.info("Detrending bck for target %s with %s points with duration %s", object_id, series_len, transit_duration * 4)
        bck_flux_detrended = np.array(lc_data.background_flux)
        bck_flux_detrend = detrend(time_values, bck_flux_detrended, transit_duration * 4, smooth=False)
        logging.info("Finished detrending for target %s with %s points", object_id, series_len)
        lc_df['bck_flux'] = bck_flux_detrend
        flux_err_values = np.array(lc_data.flux_err) / flux_values
        lc_df['#time'] = time_values
        lc_df['flux'] = flux_values
        lc_df['flux_err'] = flux_err_values
        lc_df['centroid_x'] = np.array(lc_data.centroids_x)
        lc_df['centroid_y'] = np.array(lc_data.centroids_y)
        lc_df['motion_x'] = np.array(lc_data.motion_x)
        lc_df['motion_y'] = np.array(lc_data.motion_y)
        lc_df = lc_df.fillna(0)
        lc_df.to_csv(lc_filename)
    logging.info("Computing T0s for %s", object_id)
    t0s_list = LcbuilderHelper.compute_t0s(lc_data['time'].to_numpy(), transit_period, epoch, transit_duration)
    source_offsets_diffimg = []
    source_offsets_bls = []
    centroids_offsets_ra_list = []
    centroids_offsets_dec_list = []
    centroids_offsets_time_list = []
    og_time_list = []
    og_core_flux_list = []
    og_halo_flux_list = []
    logging.info("Computing TPFs data for %s", object_id)
    if tpfs is not None:
        for tpf in tpfs:
            sector_name, sector = LcbuilderHelper.mission_lightkurve_sector_extraction(mission, tpf)
            logging.info("Computing TPF data for %s sector %s", object_id, sector)
            aperture = tpf.pipeline_mask
            cadence_s = np.round(np.nanmedian(tpf.time.value[1:] - tpf.time.value[:-1]) * 3600 * 24)
            cadences_per_transit = LcbuilderHelper.estimate_transit_cadences(cadence_s, transit_duration * 2)
            t0s_in_tpf_indexes = np.argwhere((t0s_list > tpf.time.value[0] - transit_duration) &
                                             (t0s_list < tpf.time.value[-1] + transit_duration)).flatten()
            if len(t0s_in_tpf_indexes) == 0:
                continue
            tpf_t0s_list = t0s_list[t0s_in_tpf_indexes]
            good_quality_t0s = []
            for t0 in tpf_t0s_list:
                t0_in_tpf_indexes = \
                    np.argwhere((tpf.time.value > t0 - transit_duration) & (tpf.time.value < t0 + transit_duration)).flatten()
                cadences_ratio = len(t0_in_tpf_indexes) / cadences_per_transit
                if cadences_ratio >= 0.75:
                    good_quality_t0s.append(t0)
                else:
                    try:
                        tpf = tpf[(tpf.time.value < t0 - transit_duration) | (tpf.time.value > t0 + transit_duration)]
                        trash = tpf[0] #we try to make it crash and if so, we discard the TPF
                    except:
                        logging.exception("TPF filtering error for %s and T0 %.02f", object_id, t0)
                        continue
            if len(good_quality_t0s) == 0:
                continue
            hdu = tpf.hdu[2].header
            wcs = WCS(hdu)
            if 'all' in modes or 'centroids' in modes:
                logging.info("Computing TPF data centroids for %s sector %s", object_id, sector)
                centroids_drift = \
                    Watson.compute_centroids_for_tpf(ra, dec, lc_data, tpf, wcs, transit_period, epoch, duration_to_period)
                centroids_offsets_time_list.append(centroids_drift[0])
                centroids_offsets_ra_list.append(centroids_drift[1])
                centroids_offsets_dec_list.append(centroids_drift[2])
            if 'all' in modes or 'og' in modes:
                logging.info("Computing TPF data og for %s sector %s", object_id, sector)
                optical_ghost_data = Watson.compute_optical_ghost_data(tpf, aperture, transit_period, epoch, transit_duration)
                og_time_list.append(optical_ghost_data[0])
                og_core_flux_list.append(optical_ghost_data[1])
                og_halo_flux_list.append(optical_ghost_data[2])
            if 'all' in modes or 'source_offset' in modes:
                pixel_values_i = np.array(range(tpf.shape[1]))
                pixel_values_j = np.array(range(tpf.shape[2]))
                logging.info("Computing TPF data per-pixel BLS for %s sector %s", object_id, sector)
                snr_map, ax = Watson.plot_pixels(tpf, title=mission_prefix + ' ' + str(id) + ' ' + sector_name + ' ' + str(sector) + ' TPF BLS Analysis',
                                                 period=transit_period, epoch=epoch, duration=transit_duration, aperture_mask=aperture, dry=True)
                source_offset_px = Watson.light_centroid(snr_map, pixel_values_i, pixel_values_j)
                source_offset_bls = wcs.pixel_to_world(source_offset_px[1], source_offset_px[0])
                source_offsets_bls.append((source_offset_bls.ra.value, source_offset_bls.dec.value))
                logging.info("Computing TPF data differential image for %s sector %s", object_id, sector)
                tpf_sub = Watson.compute_tpf_diff_image(tpf, transit_period, epoch, transit_duration)
                source_offset_diffimg_px = Watson.light_centroid(tpf_sub, pixel_values_i, pixel_values_j)
                source_offset_diffimg = wcs.pixel_to_world(source_offset_diffimg_px[1], source_offset_diffimg_px[0])
                source_offsets_diffimg.append((source_offset_diffimg.ra.value, source_offset_diffimg.dec.value))
        if 'all' in modes or 'og' in modes:
            og_df = pd.DataFrame(columns=['time', 'core_flux', 'halo_flux', 'og_flux'])
            og_df['time'] = list(chain.from_iterable(og_time_list))
            og_df['core_flux'] = list(chain.from_iterable(og_core_flux_list))
            og_df['halo_flux'] = list(chain.from_iterable(og_halo_flux_list))
            og_df['og_flux'] = og_df['halo_flux'] - og_df['core_flux']
            og_df.fillna(1, inplace=True)
            og_df.to_csv(file_prefix + '_og_' + str(cadence_name) + '.csv')
        if 'all' in modes or 'centroids' in modes:
            centroid_coords_df = pd.DataFrame(columns=['time', 'centroids_ra', 'centroids_dec'])
            centroid_coords_df['centroids_ra'] = list(chain.from_iterable(centroids_offsets_ra_list))
            centroid_coords_df['centroids_dec'] = list(chain.from_iterable(centroids_offsets_dec_list))
            centroid_coords_df['time'] = list(chain.from_iterable(centroids_offsets_time_list))
            centroid_coords_df.fillna(0, inplace=True)
            centroid_coords_df.to_csv(file_prefix + '_cent_' + str(cadence_name) + '.csv')
        if 'all' in modes or 'source_offset' in modes:
            if len(source_offsets_diffimg) > 0:
                source_offset_diffimg_ra = np.nanmedian(np.array(source_offsets_diffimg)[:, 0])
                source_offset_diffimg_dec = np.nanmedian(np.array(source_offsets_diffimg)[:, 1])
                source_offset_diffimg_ra_err = np.nanstd(np.array(source_offsets_diffimg)[:, 0])
                source_offset_diffimg_dec_err = np.nanstd(np.array(source_offsets_diffimg)[:, 1])
            else:
                source_offset_diffimg_ra = ra
                source_offset_diffimg_dec = dec
                source_offset_diffimg_ra_err = 0
                source_offset_diffimg_dec_err = 0
            if len(source_offsets_bls) > 0:
                source_offset_bls_ra = np.nanmedian(np.array(source_offsets_bls)[:, 0])
                source_offset_bls_dec = np.nanmedian(np.array(source_offsets_bls)[:, 1])
                source_offset_bls_ra_err = np.nanstd(np.array(source_offsets_bls)[:, 0])
                source_offset_bls_dec_err = np.nanstd(np.array(source_offsets_bls)[:, 1])
            else:
                source_offset_bls_ra = ra
                source_offset_bls_dec = dec
                source_offset_bls_ra_err = 0
                source_offset_bls_dec_err = 0
            offset_ra = np.mean([source_offset_bls_ra, source_offset_diffimg_ra])
            offset_dec = np.mean([source_offset_bls_dec, source_offset_diffimg_dec])
            offset_ra_err = 1/2 * np.sqrt(source_offset_bls_ra_err ** 2 + source_offset_diffimg_ra_err ** 2)
            offset_dec_err = 1/2 * np.sqrt(source_offset_bls_dec_err ** 2 + source_offset_diffimg_dec_err ** 2)
            if np.isnan(offset_ra_err) or offset_ra_err == 0.0:
                offset_ra_err = np.nanstd([source_offset_bls_ra, source_offset_diffimg_ra])
            if np.isnan(offset_dec_err) or offset_dec_err == 0.0:
                offset_dec_err = np.nanstd([source_offset_bls_dec, source_offset_diffimg_dec])
            offsets_df = pd.DataFrame(columns=['name', 'ra', 'dec', 'ra_err', 'dec_err'])
            offsets_df = offsets_df.append({'name': 'diff_img', 'ra': source_offset_diffimg_ra, 'dec': source_offset_diffimg_dec,
                               'ra_err': source_offset_diffimg_ra_err, 'dec_err': source_offset_diffimg_dec_err}, ignore_index=True)
            offsets_df = offsets_df.append({'name': 'px_bls', 'ra': source_offset_bls_ra, 'dec': source_offset_bls_dec,
                               'ra_err': source_offset_bls_ra_err, 'dec_err': source_offset_bls_dec_err}, ignore_index=True)
            offsets_df = offsets_df.append({'name': 'mean', 'ra': offset_ra, 'dec': offset_dec,
                               'ra_err': offset_ra_err, 'dec_err': offset_dec_err}, ignore_index=True)
            offsets_df.to_csv(file_prefix + '_offset_' + str(cadence_name) + '.csv')
    if 'all' in modes or 'metrics' in modes:
        logging.info("Computing bootstrap fap for %s", object_id)
        if lc_df is None:
            lc_df = pd.read_csv(lc_filename)
        if np.isnan(star_info.mass):
            star_info.mass = 1
        if np.isnan(star_info.mass_min):
            star_info.mass_min = star_info.mass - star_info.mass / 100
        if np.isnan(star_info.mass_max):
            star_info.mass_max = star_info.mass + star_info.mass / 100
        if np.isnan(star_info.radius):
            star_info.radius = 1
        if np.isnan(star_info.radius_min):
            star_info.radius_min = star_info.radius - star_info.radius / 100
        if np.isnan(star_info.radius_max):
            star_info.radius_max = star_info.radius + star_info.radius / 100
        if np.isnan(star_info.teff):
            star_info.teff = 6000
        lc_df = lc_df.loc[(~lc_df['flux_0'].isna()) & (lc_df['flux_0'] > 0)]
        bootstrap_fap = Watson.compute_bootstrap_fap(lc_df['#time'].to_numpy(), lc_df['flux_0'].to_numpy(), transit_period, transit_duration, star_info, flux_err=None)
        logging.info("Computing metrics for %s", object_id)
        lc_df['time_folded'] = foldedleastsquares.fold(lc_df['#time'].to_numpy(), transit_period, epoch + transit_period / 2)
        lc_df['time_folded_sec'] = foldedleastsquares.fold(lc_df['#time'].to_numpy(), transit_period, epoch)
        lc_df_it = lc_df[(lc_df['time_folded'] > 0.5 - duration_to_period / 2) & (lc_df['time_folded'] < 0.5 + duration_to_period / 2)]
        lc_df_it = lc_df_it.sort_values(by=['time_folded'])
        lc_df_secit = lc_df[(lc_df['time_folded_sec'] > 0.5 - duration_to_period / 2) & (lc_df['time_folded_sec'] < 0.5 + duration_to_period / 2)]
        lc_df_secit = lc_df_secit.sort_values(by=['time_folded_sec'])
        sec_depth = 1 - lc_df_secit['flux_0'].dropna().median()
        sec_depth_err = lc_df_secit['flux_0'].dropna().std()
        depth = 1 - lc_df_it['flux_0'].dropna().median()
        depth_err = lc_df_it['flux_0'].dropna().std()
        if depth <= 0:
            depth = 1 - lc_df_it['flux_0'].dropna().min()
        if sec_depth <= 0:
            sec_depth = 1 - lc_df_secit['flux_0'].dropna().min()
        rad_p = (ufloat(depth, depth_err) * (ufloat(star_info.radius, np.nanmax([star_info.radius - star_info.radius_min, star_info.radius_max - star_info.radius])) ** 2)) ** 0.5
        rp = rad_p.n
        rp_err = rad_p.s
        rp = LcbuilderHelper.convert_from_to(rp, u.R_sun, u.R_earth)
        habitability_calculator = HabitabilityCalculator()
        planet_eq_temp, planet_eq_temp_low_err, planet_eq_temp_up_err = (
            habitability_calculator.calculate_teq(star_info.mass, star_info.mass - star_info.mass_min, star_info.mass_max - star_info.mass,
                                                  star_info.radius, star_info.radius - star_info.radius_min, star_info.radius_max - star_info.radius,
                                                  transit_period, 0.0001, 0.0001, star_info.teff, star_info.teff / 100, star_info.teff / 100,
                                                  albedo=0.3))
        planet_eff_temp, planet_eff_temp_low_err, planet_eff_temp_up_err = (
            habitability_calculator.calculate_teff(star_info.teff, star_info.teff / 100, star_info.teff / 100,
                                                   sec_depth, sec_depth_err, sec_depth_err, depth, depth_err, depth_err))
        albedo, albedo_low_err, albedo_up_err = (
            habitability_calculator.calculate_albedo(sec_depth, sec_depth_err, sec_depth_err,
                                                     transit_period, 0.0001, 0.0001,
                                                     star_info.mass, star_info.mass - star_info.mass_min, star_info.mass_max - star_info.mass,
                                                     rp, rp_err, rp_err))
        temp_stat = habitability_calculator.calculate_planet_temperature_stat(
            planet_eq_temp, planet_eq_temp_low_err, planet_eq_temp_up_err,
            planet_eff_temp, planet_eff_temp_low_err, planet_eff_temp_up_err
        )
        albedo_stat = habitability_calculator.calculate_albedo_stat(albedo, albedo_low_err, albedo_up_err)
        statistics_df = pd.DataFrame(columns=['name', 'value'])
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['depth'], 'value': [depth]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['depth_err'], 'value': [depth_err]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['bootstrap_fap'], 'value': [bootstrap_fap]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['sec_depth'], 'value': [sec_depth]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['sec_depth_err'], 'value': [sec_depth_err]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['teff'], 'value': [planet_eff_temp]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['teff_err'], 'value': [np.nanmax([planet_eff_temp_low_err, planet_eff_temp_up_err])]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['teq'], 'value': [planet_eq_temp]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['teq_err'], 'value': [np.nanmax([planet_eq_temp_low_err, planet_eq_temp_up_err])]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['temp_stat'], 'value': [temp_stat]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['geom_albedo'], 'value': [albedo]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['geom_albedo_err'], 'value': [np.nanmax([albedo_low_err, albedo_up_err])]}, orient='columns')], ignore_index=True)
        statistics_df = pd.concat([statistics_df, pd.DataFrame.from_dict({'name': ['albedo_stat'], 'value': [albedo_stat]}, orient='columns')], ignore_index=True)
        statistics_df.to_csv(file_prefix + '_metrics_' + str(cadence_name) + '.csv')
    return lc_df

class CreateTargetInput:
    def __init__(self, mission, id, lc_file, target_dir, models_dir, injected_files_df, cache_dir, index,
                 max_lc_length=20610, force=False, tpf_files=None, inject_random=False, mode='all') -> None:
        super().__init__()
        self.mission = mission
        self.id = id
        self.lc_file = lc_file
        self.target_dir = target_dir
        self.models_dir = models_dir
        self.injected_files_df = injected_files_df
        self.max_lc_length = max_lc_length
        self.cache_dir = cache_dir
        self.index = index
        self.force = force
        self.tpf_files = tpf_files
        self.inject_random = inject_random
        self.mode = mode

def create_target_csv(create_target_input):
    mission_lightcurve_builder = MissionLightcurveBuilder()
    mission_input_lightcurve_builder = MissionInputLightcurveBuilder()
    mission_id = create_target_input.id
    file_id = mission_id.replace(' ', '_')
    file_prefix = create_target_input.target_dir + file_id
    file_planet_prefix = file_prefix + '_' + \
                         str(round(create_target_input.injected_files_df['period'], 2))
    pid = multiprocessing.current_process().pid
    tpfs_dir = create_target_input.target_dir + 'tpfs/'
    sectors = 'all'
    if 'Sectors' in create_target_input.injected_files_df:
        sectors = [int(item) for item in create_target_input.injected_files_df['Sectors'].split(",")]
    elif 'sectors' in create_target_input.injected_files_df:
        sectors = [int(create_target_input.injected_files_df['sectors'])] if create_target_input.injected_files_df['sectors'].isdigit() else 'all'
    try:
        logging.info("PID %s: Trying to get short cadence info for %s", pid, file_planet_prefix)
        file_name = file_planet_prefix + '_metrics_short.csv'
        authors = [constants.KEPLER_AUTHOR] if mission_id.startswith('KIC') else [constants.SPOC_AUTHOR]
        cadences = [60] if mission_id.startswith('KIC') else [120]
        if create_target_input.force or not os.path.exists(file_name):
            if create_target_input.lc_file is None:
                lcbuild_short = \
                    mission_lightcurve_builder.build(MissionObjectInfo(sectors,
                                                                       mission_id, author=authors,
                                                                       cadence=cadences,
                                                                       high_rms_enabled=False),
                                                                       create_target_input.target_dir,
                                                                       create_target_input.cache_dir)
                if os.path.exists(tpfs_dir):
                    for f in glob.glob(tpfs_dir + f"*{mission_id}*.fits"):
                        os.remove(f)
            else:
                if isinstance(create_target_input.lc_file, (list, np.ndarray)):
                    lc = None
                    for file in create_target_input.lc_file:
                        if lc is None:
                            lc = KeplerLightCurve.read(file)
                            csv_filename = os.path.splitext(file)[0] + '.csv'
                        else:
                            lc.append(KeplerLightCurve.read(file))
                else:
                    lc = KeplerLightCurve.read(create_target_input.lc_file)
                    csv_filename = os.path.splitext(create_target_input.lc_file)[0] + '.csv'
                lc_data = mission_input_lightcurve_builder.extract_lc_data(LightCurveCollection([lc]))
                lc_data['#time'] = lc_data['time']
                lc_data.to_csv(csv_filename, index=False)
                lcbuild_short = \
                    mission_input_lightcurve_builder.build(
                        MissionInputObjectInfo(mission_id, input_file=csv_filename, high_rms_enabled=False),
                        create_target_input.target_dir,
                        create_target_input.cache_dir)
            prepare_lc_data(file_planet_prefix, mission_id, lcbuild_short.lc_data,
                            create_target_input.injected_files_df['duration(h)'] / 24,
                            create_target_input.injected_files_df['period'],
                            create_target_input.injected_files_df['epoch'],
                            lcbuild_short.star_info.ra, lcbuild_short.star_info.dec, lcbuild_short.star_info,
                            sectors=sectors, authors=authors, cadences=cadences, cadence_name='short',
                            modes=create_target_input.mode)
            for f in pathlib.Path(tpfs_dir).glob(f"*{file_id.split('_')[1]}*"):
                os.remove(f)
        logging.info("PID %s: Finished to get short cadence info for %s", pid, file_planet_prefix)
    except Exception as e:
        logging.exception("PID %s: Failed short cadence for %s", pid, file_planet_prefix)
    logging.info(f"lc file is {create_target_input.lc_file}")
    if create_target_input.lc_file is None:
        authors = [constants.KEPLER_AUTHOR] if mission_id.startswith('KIC') else [constants.SPOC_AUTHOR, constants.TESS_SPOC_AUTHOR, 'QLP']
        cadences = [1800] if mission_id.startswith('KIC') else [200, 600, 1800]
        try:
            logging.info("PID %s: Trying to get long cadence info for %s", pid, file_planet_prefix)
            file_name = file_planet_prefix + '_metrics_long.csv'
            if create_target_input.force or not os.path.exists(file_name):
                lcbuild_long = \
                    mission_lightcurve_builder.build(MissionObjectInfo(sectors,
                                            mission_id, author=authors, cadence=cadences, high_rms_enabled=False),
                                            create_target_input.target_dir, create_target_input.cache_dir)
                prepare_lc_data(file_planet_prefix, mission_id, lcbuild_long.lc_data,
                                create_target_input.injected_files_df['duration(h)'] / 24,
                                create_target_input.injected_files_df['period'],
                                create_target_input.injected_files_df['epoch'],
                                lcbuild_long.star_info.ra, lcbuild_long.star_info.dec, lcbuild_long.star_info,
                                sectors=sectors, authors=authors, cadences=cadences, cadence_name='long',
                                modes=create_target_input.mode)
                for f in pathlib.Path(tpfs_dir).glob(f"*{file_id.split('_')[1]}*"):
                    os.remove(f)
            logging.info("PID %s: Finished to get long cadence info for index %s for file %s", pid,
                         create_target_input.index, file_planet_prefix)
        except Exception as e:
            logging.exception("PID %s: Failed long cadence for %s", pid, file_planet_prefix)

def create_triceratops_prob(create_target_input):
    object_id = create_target_input.id
    file_id = object_id.replace(' ', '_')
    file_prefix = create_target_input.target_dir + file_id
    file_planet_prefix = file_prefix + '_' + \
                         str(round(create_target_input.injected_files_df['period'], 2))
    logging.exception(f"Trying triceratops for target {file_planet_prefix}")
    triceratops_file = file_planet_prefix + '_triceratops.csv'
    if os.path.exists(triceratops_file) and not create_target_input.force:
        return
    try:
        period = create_target_input.injected_files_df['period']
        duration = create_target_input.injected_files_df['duration(h)']
        epoch = create_target_input.injected_files_df['epoch']
        depth = create_target_input.injected_files_df['depth'] / 1e6
        mission, mission_prefix, id_int = LcBuilder().parse_object_info(object_id)
        apertures = {}
        lc_short_filename = str(file_prefix) + '_lc_short.csv'
        lc_long_filename = str(file_prefix) + '_lc_long.csv'
        if create_target_input.tpf_files is None:
            if os.path.exists(lc_short_filename):
                tpfs = lightkurve.search_targetpixelfile(object_id, exptime="short", author=lcbuilder.constants.KEPLER_AUTHOR).download_all()
            elif os.path.exists(lc_long_filename):
                tpfs = lightkurve.search_targetpixelfile(object_id, exptime="long", author=lcbuilder.constants.KEPLER_AUTHOR).download_all()
        else:
            tpfs = []
            for file in create_target_input.tpf_files:
                tpfs.append(lightkurve.TessTargetPixelFile(file, ignore_missing_simple=True))
        sectors = []
        for tpf in tpfs:
            if mission_prefix == constants.MISSION_ID_KEPLER:
                sector = tpf.quarter
            elif mission_prefix == constants.MISSION_ID_TESS:
                sector = tpf.sector
            if mission_prefix == constants.MISSION_ID_KEPLER_2:
                sector = tpf.campaign
            sectors = sectors + [sector]
            apertures[sector] = ApertureExtractor.from_boolean_mask(tpf.pipeline_mask, tpf.column, tpf.row)
        sectors = np.array(sectors)
        logging.info(f"Will execute validation for sectors: {sectors}")
        lc_df = None
        if os.path.exists(lc_short_filename):
            lc_df = pd.read_csv(lc_short_filename, header=0)
        if os.path.exists(lc_long_filename):
            if lc_df is None:
                lc_df = pd.read_csv(lc_long_filename, header=0)
            else:
                lc_df = pd.concat([lc_df, pd.read_csv(lc_long_filename, header=0)])
        lc_df['time_folded'] = foldedleastsquares.core.fold(lc_df['#time'].to_numpy(), period, epoch + period / 2)
        range = 5
        bins = 200
        lc_df = lc_df.sort_values(by=['time_folded'], ascending=True)
        half_duration_phase = duration / 2 / period
        folded_plot_range = half_duration_phase * range
        folded_plot_range = folded_plot_range if folded_plot_range < 0.5 else 0.5
        lc_df = lc_df.loc[(lc_df['time_folded'] > 0.5 - folded_plot_range) &
                          (lc_df['time_folded'] < 0.5 + folded_plot_range)]
        bin_centers, bin_means, bin_width, bin_stds = \
            LcbuilderHelper.bin(lc_df['time_folded'].to_numpy(), lc_df['flux'].to_numpy(), bins,
                                values_err=lc_df['flux_err'].to_numpy(), bin_err_mode='values_snr')
        bin_centers = (bin_centers - 0.5) * period
        sigma = np.nanmean(bin_stds)
        logging.info("Flux err (ppm) = %s", sigma * 1000000)
        logging.info("Acquiring triceratops target")
        target = tr.target(ID=id_int, mission=mission, sectors=sectors)
        # TODO allow user input apertures
        logging.info("Reading apertures from directory")
        logging.info("Calculating validation closest stars depths")
        target.calc_depths(depth, apertures)
        logging.info("Preparing validation processes inputs")
        target.calc_probs(time=bin_centers, flux_0=bin_means, flux_err_0=sigma, P_orb=float(period),
                          contrast_curve_file=None, parallel=True)
        target.probs.to_csv(triceratops_file, index=False)
    except Exception as e:
        logging.exception(f"Failed triceratops for target {file_planet_prefix}")

    #input.target.probs
    # scenario  prob
    # group by scenario with sum of probs
    # FPP
    # NFPP
    # TP
    # EB
    # EBx2P
    # PTP
    # PEB
    # PEBx2P
    # STP
    # SEB
    # SEBx2P
    # DTP
    # DEB
    # DEBx2P
    # BTP
    # BEB
    # BEBx2P
    # NTP other stars
    # NEB other stars
    # NEBx2P other stars

    # input.target.plot_fits(save=True, fname=input.save_dir + "/scenario_" + str(input.run) + "_fits",
    #                        time=input.time, flux_0=input.flux, flux_err_0=input.sigma)
    # return input.target.FPP, input.target.NFPP, fpp2, fpp3, input.target.probs, input.target.star_num, \
    #     input.target.u1, input.target.u2, input.target.fluxratio_EB, input.target.fluxratio_comp



from keras.losses import BinaryCrossentropy
from keras.metrics import BinaryAccuracy

from exoml.ml.loss.cross_entropy import binary_focal_loss
from exoml.ml.model.binary_model import BinaryModel


class ImbalancedBinaryModel(BinaryModel):
    def __init__(self, name, input_size, class_ids, type_to_label, hyperparams) -> None:
        super().__init__(name, input_size, class_ids, type_to_label, hyperparams)

    def instance_loss_accuracy(self):
        return BinaryCrossentropy(), BinaryAccuracy()
